import logging
import sys

from dsw2to3.consts import LOGGER_NAME


logging.basicConfig(
    stream=sys.stdout,
    level=logging.INFO,
    format='%(asctime)s | %(levelname)8s | %(message)s',
)
LOGGER = logging.getLogger(LOGGER_NAME)



import dataclasses

from dsw2to3.config import Config


def insert_query(table_name: str, fields: list[str]):
    field_names = ', '.join(fields)
    field_placeholders = ', '.join('%s' for _ in fields)
    return f'INSERT INTO {table_name} ({field_names}) VALUES ({field_placeholders});'


def validate_mongo_doc(document: dict, required_fields: list[str]):
    missing = [field for field in required_fields
               if field not in document.keys()]
    if len(missing) > 0:
        raise RuntimeError(f'Missing fields: {missing}')


@dataclasses.dataclass
class MigrationOptions:
    dry_run: bool
    skip_pre_check: bool
    fix_integrity: bool


class Migrator:

    def __init__(self, config: Config, options: MigrationOptions):
        self.config = config
        self.options = options

    def pre_check(self) -> bool:
        pass

    def migrate(self):
        pass

    def finish(self):
        pass



import psycopg2
import psycopg2.extensions
import tenacity

from dsw2to3.config import PostgresConfig
from dsw2to3.errors import ERROR_HANDLER
from dsw2to3.logger import LOGGER


class PostgresDB:

    def __init__(self, config: PostgresConfig):
        self.config = config
        self.dsn = psycopg2.extensions.make_dsn(
            dsn=config.connection_string
        )
        self._connection = None

    @tenacity.retry(
        reraise=True,
        wait=tenacity.wait_exponential(multiplier=0.3),
        stop=tenacity.stop_after_attempt(3),
    )
    def _connect_db(self):
        LOGGER.debug(f'Creating connection to PostgreSQL database')
        connection = psycopg2.connect(dsn=self.dsn)
        # test connection
        with connection.cursor() as cursor:
            cursor.execute(query='SELECT 1;')
        connection.commit()
        self._connection = connection

    def connect(self):
        if not self._connection or self._connection.closed != 0:
            self._connect_db()

    @property
    def connection(self):
        self.connect()
        return self._connection

    def new_cursor(self):
        return self.connection.cursor()

    def reset(self):
        self.close()
        self.connect()

    def commit(self):
        self.connection.commit()

    def close(self):
        if self._connection:
            LOGGER.info(f'Closing connection to PostgreSQL database')
            self._connection.close()
        self._connection = None

    def execute(self, entity, instance):
        LOGGER.info(f'Executing INSERT INTO {entity.TABLE_NAME}'
                    f' (single)')
        try:
            with self.new_cursor() as cursor:
                cursor.execute(
                    query=entity.INSERT_QUERY,
                    vars=instance.query_vars(),
                )
        except Exception as e:
            ERROR_HANDLER.error(
                cause='PostgreSQL',
                message=f'Failed to insert {entity.__name__} ({e})'
                        f' with vars {instance.query_vars()}',
            )

    def execute_loop(self, entity, instances):
        with self.new_cursor() as cursor:
            for instance in instances:
                try:
                    cursor.execute(
                        query=entity.INSERT_QUERY,
                        vars=instance.query_vars(),
                    )
                except Exception as e:
                    ERROR_HANDLER.error(
                        cause='PostgreSQL',
                        message=f'Failed to insert {entity.__name__} ({e})'
                                f' with vars {instance.query_vars()}',
                    )

    def disable_triggers(self, table: str):
        with self.new_cursor() as cursor:
            cursor.execute(f'ALTER TABLE {table} DISABLE TRIGGER ALL;')

    def enable_triggers(self, table: str):
        with self.new_cursor() as cursor:
            cursor.execute(f'ALTER TABLE {table} ENABLE TRIGGER ALL;')



import datetime
import gridfs
import pymongo
import tenacity

from typing import Optional

from dsw2to3.config import MongoConfig
from dsw2to3.errors import ERROR_HANDLER
from dsw2to3.logger import LOGGER

DOCUMENT_FS_COLLECTION = 'documentFs'
ASSETS_FS_COLLECTION = 'templateAssetFs'


def _fetch_file(fs: gridfs.GridFS, file_name: str) -> Optional[bytes]:
    data = None
    try:
        file = fs.find_one({'filename': file_name})
        if file is not None:
            data = file.read()
            file.close()
    except Exception as e:
        LOGGER.debug(f'Failed to retrieve file from GridFS '
                     f'with filename {file_name}: {e}')
    return data


class MongoDB:

    def __init__(self, config: MongoConfig):
        self.config = config
        self.client = pymongo.MongoClient(**self.config.mongo_client_kwargs)
        self.db = self.client[self.config.database]
        self.doc_fs = gridfs.GridFS(self.db, DOCUMENT_FS_COLLECTION)
        self.asset_fs = gridfs.GridFS(self.db, ASSETS_FS_COLLECTION)
        self.now = datetime.datetime.now()

    def update_now(self):
        self.now = datetime.datetime.now()

    @tenacity.retry(
        reraise=True,
        wait=tenacity.wait_exponential(multiplier=0.5),
        stop=tenacity.stop_after_attempt(3),
    )
    def load_list(self, entity) -> list:
        result = list()
        for doc in self.db[entity.COLLECTION].find():
            try:
                result.append(entity.from_mongo(doc, self.now))
            except Exception as e:
                ERROR_HANDLER.error(
                    cause='MongoDB',
                    message=f'- cannot load {entity.__name__} '
                            f'({e}): {doc}'
                )
        LOGGER.info(f'- loaded {entity.__name__}: {len(result)} entries')
        return result

    @tenacity.retry(
        reraise=True,
        wait=tenacity.wait_exponential(multiplier=0.5),
        stop=tenacity.stop_after_attempt(3),
    )
    def load_nested(self, source_entity, target_entity, field: str):
        result = list()
        for doc in self.db[source_entity.COLLECTION].find():
            children = doc.get(field, [])
            try:
                for child in children:
                    result.append(target_entity.from_mongo(doc, child, self.now))
            except Exception as e:
                ERROR_HANDLER.error(
                    cause='MongoDB',
                    message=f'- cannot load {target_entity.__name__} '
                            f'({e}): {doc}'
                )
        LOGGER.info(f'- loaded {target_entity.__name__}: {len(result)} entries')
        return result

    @tenacity.retry(
        reraise=True,
        wait=tenacity.wait_exponential(multiplier=0.5),
        stop=tenacity.stop_after_attempt(3),
    )
    def fetch_document(self, file_name: str) -> Optional[bytes]:
        return _fetch_file(self.doc_fs, file_name)

    @tenacity.retry(
        reraise=True,
        wait=tenacity.wait_exponential(multiplier=0.5),
        stop=tenacity.stop_after_attempt(3),
    )
    def fetch_asset(self, file_name: str) -> Optional[bytes]:
        return _fetch_file(self.asset_fs, file_name)




from typing import List

from dsw2to3.config import Config
from dsw2to3.connection import PostgresDB, MongoDB, S3Storage
from dsw2to3.errors import ERROR_HANDLER
from dsw2to3.migration.common import Migrator, MigrationOptions
from dsw2to3.migration.entitites import *


class WizardMigrator(Migrator):

    # All tables but in specific order (bcs FK)
    _TABLES_CLEANUP = [
        Level.TABLE_NAME,
        Metric.TABLE_NAME,
        ActionKey.TABLE_NAME,
        AppConfig.TABLE_NAME,
        BookReference.TABLE_NAME,
        Branch.TABLE_NAME,
        'document_queue',
        Feedback.TABLE_NAME,
        KMMigration.TABLE_NAME,
        QuestionnaireMigration.TABLE_NAME,
        QuestionnaireACLGroup.TABLE_NAME,
        QuestionnaireACLUser.TABLE_NAME,
        ACLGroup.TABLE_NAME,
        TemplateAsset.TABLE_NAME,
        TemplateFile.TABLE_NAME,
        Document.TABLE_NAME,
        Questionnaire.TABLE_NAME,
        Template.TABLE_NAME,
        Package.TABLE_NAME,
        User.TABLE_NAME,
    ]

    _LOAD_ENTITIES_SIMPLE = [
        ActionKey,
        AppConfig,
        BookReference,
        Branch,
        Document,
        Feedback,
        KMMigration,
        Level,
        Metric,
        Package,
        QuestionnaireMigration,
        Questionnaire,
        Package,
        Template,
        User,
    ]

    _LOAD_ENTITIES_NESTED = [
        (QuestionnaireACLUser, Questionnaire, 'permissions'),
        (TemplateAsset, Template, 'assets'),
        (TemplateFile, Template, 'files'),
    ]

    _DEFAULT_TABLE_COUNTS = {
        ACLGroup.TABLE_NAME: 0,
        ActionKey.TABLE_NAME: 0,
        AppConfig.TABLE_NAME: 1,  # default app config
        BookReference.TABLE_NAME: 0,
        Branch.TABLE_NAME: 0,
        Document.TABLE_NAME: 0,
        'document_queue': 0,
        Feedback.TABLE_NAME: 0,
        KMMigration.TABLE_NAME: 0,
        Level.TABLE_NAME: 3,  # default levels (proposal, DMP, project)
        Metric.TABLE_NAME: 6,  # default metrics (F,A,I,R,G,O)
        Package.TABLE_NAME: 0,
        QuestionnaireMigration.TABLE_NAME: 0,
        Questionnaire.TABLE_NAME: 0,
        QuestionnaireACLGroup.TABLE_NAME: 0,
        QuestionnaireACLUser.TABLE_NAME: 0,
        Template.TABLE_NAME: 0,
        TemplateAsset.TABLE_NAME: 0,
        TemplateFile.TABLE_NAME: 0,
        User.TABLE_NAME: 3,  # default users (albert, nikola, isaac)
    }

    _INSERT_ARGS = [
        (User, False),  # User (FKs: User)
        (Package, True),  # Package (FKs: Package)
        (Template, False),  # Template (FKs: -)
        (Questionnaire, False),  # Questionnaire (FKs: Package, Template, User)
        (ActionKey, False),  # ActionKey (FKs: User)
        (AppConfig, False),  # AppConfig (FKs: -)
        (BookReference, False),  # BookReference (FKs: -)
        (Branch, False),  # Branch (FKs: Package, User)
        (Document, False),  # Document (FKs: Questionnaire, Template, User)
        (Feedback, False),  # Feedback (FKs: Package)
        (KMMigration, False),  # KMMigration (FKs: Branch, Package)
        (Level, False),  # Level (FKs: -)
        (Metric, False),  # Metric (FKs: -)
        (QuestionnaireMigration, False),  # QuestionnaireMigration (FKs: Questionnaire)
        (QuestionnaireACLUser, False),  # QuestionnaireACLUser (FKs: Questionnaire, User)
        (TemplateAsset, False),  # TemplateAsset (FKs: Template)
        (TemplateFile, False),  # TemplateFile (FKs: Template)
    ]

    def __init__(self, config: Config, options: MigrationOptions):
        super().__init__(config, options)
        self.entities = WizardEntities()
        self.postgres = PostgresDB(config=config.postgres)
        self.mongo = MongoDB(config=config.mongo)
        self.s3 = S3Storage(config=config.s3)

    def pre_check(self) -> bool:
        all_ok = True
        LOGGER.info(f'Checking source database (MongoDB)')
        try:
            source_migrations = self.mongo.db['migrations'].count_documents({})
            expect_migrations = 42  # There are 42 DB migrations in DSW 2.14.0
            if source_migrations != expect_migrations:
                all_ok = False
                ERROR_HANDLER.error(
                    cause='MongoDB',
                    message=f'- source database has incorrect version '
                            f'(has {source_migrations} migrations, '
                            f'expected {expect_migrations})',
                )
            else:
                LOGGER.info('- source database seems to has correct version')
        except Exception as e:
            all_ok = False
            ERROR_HANDLER.critical(
                cause='MongoDB',
                message=f'- failed to check data in MongoDB ({e})',
            )
        LOGGER.info(f'Checking target database (PostgreSQL)')
        try:
            with self.postgres.new_cursor() as cursor:
                cursor.execute(query='SELECT * FROM migration;')
                result = cursor.fetchall()
            if len(result) != 1:  # There is 1 DB migration in DSW 3.0.0
                all_ok = False
                ERROR_HANDLER.error(
                    cause='PostgreSQL',
                    message=f'- target database has incorrect version '
                            f'(has {len(result)} migrations, expected 1)',
                )
            else:
                LOGGER.info('- target database seems to has correct version')
            for table, count in self._DEFAULT_TABLE_COUNTS.items():
                try:
                    with self.postgres.new_cursor() as cursor:
                        cursor.execute(
                            query=f'SELECT count(*) FROM {table};',
                        )
                        result = cursor.fetchone()[0]
                        if result > count:
                            all_ok = False
                            ERROR_HANDLER.warning(
                                cause='PostgreSQL',
                                message=f'- there are some custom data in table {table}',
                            )
                except Exception as e:
                    all_ok = False
                    ERROR_HANDLER.critical(
                        cause='PostgreSQL',
                        message=f'- failed to check table "{table}" ({e})',
                    )
        except Exception as e:
            all_ok = False
            ERROR_HANDLER.critical(
                cause='PostgreSQL',
                message=f'- failed to check data in PostgreSQL ({e})',
            )
        LOGGER.info(f'Checking target file storage (S3)')
        try:
            if self.s3.bucket_exists():
                LOGGER.info(f'- S3 storage seems to be OK (bucket exists)')
            else:
                if not self._dry_run_tag():
                    self.s3.ensure_bucket()
                LOGGER.info(f'- S3 storage seems to be OK (bucket created)'
                            f'{self._dry_run_tag()}')
        except Exception as e:
            all_ok = False
            ERROR_HANDLER.critical(
                cause='S3',
                message=f'- failed to check data in S3 storage ({e})',
            )
        return all_ok

    def cleanup_postgres(self):
        LOGGER.info(f'Cleaning up target database{self._dry_run_tag()}')
        try:
            with self.postgres.new_cursor() as cursor:
                for table in self._TABLES_CLEANUP:
                    cursor.execute(
                        query=f'DELETE FROM {table};',
                    )
                    rows_deleted = cursor.rowcount
                    LOGGER.debug(f'- deleted {rows_deleted} from {table}')
                if not self.options.dry_run:
                    self.postgres.commit()
        except Exception as e:
            ERROR_HANDLER.critical(
                cause='PostgreSQL',
                message=f'- failed to clean up PostgreSQL ({e})',
            )

    def cleanup_s3(self):
        LOGGER.info(f'Checking existing data in S3 bucket'
                    f'{self._dry_run_tag()}')
        if self.s3.bucket_exists():
            documents = self.s3.count_documents()
            templates = self.s3.count_templates()
            LOGGER.debug(f'- there are {documents} documents and {templates} '
                         f'templates in the S3 bucket')
            if documents > 0 or templates > 0:
                LOGGER.info(f'- cleaning S3 bucket (deleting objects){self._dry_run_tag()}')
                if not self.options.dry_run:
                    self.s3.delete_documents()
                    self.s3.delete_templates()
        else:
            LOGGER.info(f'- creating S3 bucket{self._dry_run_tag()}')
            if not self.options.dry_run:
                self.s3.ensure_bucket()

    def load(self):
        LOGGER.info(f'Loading data from MongoDB')
        self.mongo.update_now()

        for entity in self._LOAD_ENTITIES_SIMPLE:
            self.entities.set_list(entity, self.mongo.load_list(entity=entity))

        for entity, parent_entity, field in self._LOAD_ENTITIES_NESTED:
            self.entities.set_list(entity, self.mongo.load_nested(
                source_entity=parent_entity,
                target_entity=entity,
                field=field,
            ))

    def check_integrity(self):
        LOGGER.info('Checking data integrity')
        issues = self.entities.check_integrity()
        for issue in issues:
            LOGGER.warning(f'- violation: {issue}')
        if not self.options.fix_integrity and len(issues) > 0:
            ERROR_HANDLER.error(
                cause='Integrity',
                message='- data integrity violated (see logs above). '
                        'You can skip invalid entries by re-run with '
                        'flag --fix-integrity.'
            )

    def _dry_run_tag(self):
        return ' [--dry-run]' if self.options.dry_run else ''

    def _run_insert(self, entity, disable_triggers: bool = False):
        instances = self.entities.list_by_entity(entity)
        ok_instances = [e for e in instances if e.integrity_ok is True]
        LOGGER.info(f'- executing INSERT INTO {entity.TABLE_NAME} '
                    f'({len(ok_instances)} of {len(instances)})')
        if disable_triggers:
            self.postgres.disable_triggers(entity.TABLE_NAME)
        self.postgres.execute_loop(entity=entity, instances=ok_instances)
        if disable_triggers:
            self.postgres.enable_triggers(entity.TABLE_NAME)

    def insert(self):
        LOGGER.info(f'Inserting data to target PostgreSQL '
                    f'database{self._dry_run_tag()}')

        for entity, disable_triggers in self._INSERT_ARGS:
            self._run_insert(entity=entity, disable_triggers=disable_triggers)

        try:
            LOGGER.info(f'- committing transaction{self._dry_run_tag()}')
            if not self.options.dry_run:
                self.postgres.commit()
        except Exception as e:
            ERROR_HANDLER.critical(
                cause='PostgreSQL',
                message=f'- failed to commit transaction: {e}'
            )

    def migrate_documents(self):
        counter = 0
        not_done = 0
        skipped = 0
        documents = list(self.entities.valid_entries(Document))  # type: List[Document]
        LOGGER.info(f'Migrating documents ({len(documents)}'
                    f' files){self._dry_run_tag()}')
        for document in documents:
            if document.state != 'DoneDocumentState':
                LOGGER.debug(f'- skipping migrating file for document {document.uuid},'
                             f'because of its state ({document.state})')
                not_done += 1
                continue
            data = self.mongo.fetch_document(document.uuid)
            if data is None:
                LOGGER.warning(
                    f' - no data found for document {document.uuid},'
                    f'cannot be transferred to S3 storage - skipping'
                )
                skipped += 1
                continue
            if not self.options.dry_run:
                self.s3.store_document(
                    file_name=document.uuid,
                    content_type=document.metadata.get(
                        'contentType',
                        'application/octet-stream',
                    ),
                    data=data,
                )
            counter += 1
        LOGGER.info(f'- {counter} documents stored in S3 '
                    f'({not_done} not in "done" state, '
                    f'{skipped} not found and skipped)')

    def migrate_template_assets(self):
        counter = 0
        skipped = 0
        assets = list(self.entities.valid_entries(TemplateAsset))  # type: List[TemplateAsset]
        LOGGER.info(f'Migrating assets ({len(assets)} '
                    f'assets){self._dry_run_tag()}')
        for asset in assets:
            data = self.mongo.fetch_asset(asset.original_uuid)
            if data is None:
                LOGGER.warning(
                    f' - no data found for asset {asset.uuid} '
                    f'(original UUID {asset.original_uuid}),'
                    f'cannot be transferred to S3 storage - skipping'
                )
                skipped += 1
                continue
            if not self.options.dry_run:
                self.s3.store_template_asset(
                    template_id=asset.template_id,
                    file_name=asset.uuid,
                    content_type=asset.content_type,
                    data=data
                )
            counter += 1
        LOGGER.info(f'- {counter} template assets stored in S3 '
                    f'({skipped} not found and skipped)')

    def finish(self):
        LOGGER.debug('Closing connections')
        self.postgres.close()

    def migrate(self):
        if not self.options.skip_pre_check:
            self.pre_check()
        else:
            LOGGER.info('Skipping pre-check phase [--skip-pre-check]')

        LOGGER.info('Starting to clean up migration targets')
        self.cleanup_postgres()
        self.cleanup_s3()
        LOGGER.info('Cleaning up migration targets finished')

        LOGGER.info('Starting Wizard database migration')
        self.load()
        self.check_integrity()
        self.insert()
        LOGGER.info('Wizard database migration finished')

        LOGGER.info('Starting Wizard file storage migration')
        self.migrate_documents()
        self.migrate_template_assets()
        LOGGER.info('Wizard file storage migration finished')



import click

from typing import TextIO

from dsw2to3.config import Config, ConfigParser, MissingConfigurationError
from dsw2to3.consts import PROG_NAME, VERSION
from dsw2to3.errors import ERROR_HANDLER
from dsw2to3.logger import LOGGER
from dsw2to3.migration import WizardMigrator, MigrationOptions

LOG_LEVELS = ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']
DEFAULT_LOG_LEVEL = 'INFO'


def validate_config(ctx, param, value: TextIO) -> Config:
    content = value.read()
    cfg_parser = ConfigParser()

    if not cfg_parser.can_read(content):
        click.echo('Error: Cannot parse config file', err=True)
        exit(1)

    try:
        cfg_parser.read_string(content)
        cfg_parser.validate()
        return cfg_parser.config
    except MissingConfigurationError as e:
        click.echo('Error: Missing configuration', err=True)
        for missing_item in e.missing:
            click.echo(f' - {missing_item}')
        exit(1)


@click.command(name=PROG_NAME)
@click.option('-c', '--config', type=click.File('r'),
              callback=validate_config, required=True,
              help='Config YML file (see example).')
@click.option('-b', '--best-effort', is_flag=True,
              help='Run in best effort mode (continue on errors, not recommended).')
@click.option('-d', '--dry-run', is_flag=True,
              help='Run without making any changes in the migration targets.')
@click.option('-f', '--fix-integrity', is_flag=True,
              help='Fix integrity by skipping invalid or duplicate entries.')
@click.option('-s', '--skip-pre-check', is_flag=True,
              help='Skip initial checks of existing data in databases.')
@click.option('-l', '--log-level', type=click.Choice(LOG_LEVELS),
              default=DEFAULT_LOG_LEVEL, show_default=True,
              help='Set logging level.')
@click.version_option(version=VERSION, prog_name=PROG_NAME)
def main(config: Config, dry_run: bool, best_effort: bool, skip_pre_check: bool,
         fix_integrity: bool, log_level: str):
    """CLI tool to support data migration from DSW 2.14 to DSW 3.0"""
    LOGGER.setLevel(log_level)
    if best_effort:
        ERROR_HANDLER.set_log()
    options = MigrationOptions(
        dry_run=dry_run,
        fix_integrity=fix_integrity,
        skip_pre_check=skip_pre_check,
    )
    migrator = WizardMigrator(
        config=config,
        options=options,
    )
    migrator.migrate()
    migrator.finish()



import contextlib
import minio
import minio.datatypes
import tempfile
import tenacity

from typing import Iterable

from dsw2to3.config import S3Config


DOCUMENTS_DIR = 'documents'
TEMPLATES_DIR = 'templates'


@contextlib.contextmanager
def temp_binary_file(data: bytes):
    file = tempfile.TemporaryFile()
    file.write(data)
    file.seek(0)
    yield file
    file.close()


class S3Storage:

    @staticmethod
    def _get_endpoint(url: str):
        parts = url.split('://', maxsplit=1)
        return parts[0] if len(parts) == 1 else parts[1]

    def __init__(self, config: S3Config):
        self.config = config
        endpoint = self._get_endpoint(self.config.url)
        self.client = minio.Minio(
            endpoint=endpoint,
            access_key=self.config.username,
            secret_key=self.config.password,
            secure=self.config.url.startswith('https://'),
        )

    @property
    def identification(self) -> str:
        return f'{self.config.url}/{self.config.bucket}'

    @tenacity.retry(
        reraise=True,
        wait=tenacity.wait_exponential(multiplier=0.5),
        stop=tenacity.stop_after_attempt(3),
    )
    def bucket_exists(self):
        return self.client.bucket_exists(self.config.bucket)

    @tenacity.retry(
        reraise=True,
        wait=tenacity.wait_exponential(multiplier=0.5),
        stop=tenacity.stop_after_attempt(3),
    )
    def ensure_bucket(self):
        found = self.client.bucket_exists(self.config.bucket)
        if not found:
            self.client.make_bucket(self.config.bucket)

    @tenacity.retry(
        reraise=True,
        wait=tenacity.wait_exponential(multiplier=0.5),
        stop=tenacity.stop_after_attempt(3),
    )
    def _store_file(self, file_name: str, content_type: str, data: bytes):
        with temp_binary_file(data=data) as file:
            self.client.put_object(
                bucket_name=self.config.bucket,
                object_name=file_name,
                data=file,
                length=len(data),
                content_type=content_type,
            )

    def store_document(self, file_name: str, content_type: str, data: bytes):
        self._store_file(
            file_name=f'{DOCUMENTS_DIR}/{file_name}',
            data=data,
            content_type=content_type,
        )

    def store_template_asset(self, template_id: str, file_name: str, content_type: str, data: bytes):
        self._store_file(
            file_name=f'{TEMPLATES_DIR}/{template_id}/{file_name}',
            data=data,
            content_type=content_type,
        )

    def list_documents(self) -> Iterable[minio.datatypes.Object]:
        return self.client.list_objects(
            bucket_name=self.config.bucket,
            prefix=DOCUMENTS_DIR,
        )

    def list_templates(self, recursive=False) -> Iterable[minio.datatypes.Object]:
        return self.client.list_objects(
            bucket_name=self.config.bucket,
            prefix=TEMPLATES_DIR,
            recursive=recursive,
        )

    def count_documents(self) -> int:
        cnt = 0
        for _ in self.list_documents():
            cnt += 1
        return cnt

    def count_templates(self) -> int:
        cnt = 0
        for _ in self.list_templates():
            cnt += 1
        return cnt

    def delete_documents(self):
        for obj in self.list_documents():
            self.client.remove_object(
                bucket_name=obj.bucket_name,
                object_name=obj.object_name
            )

    def delete_templates(self):
        for obj in self.list_templates(recursive=True):
            self.client.remove_object(
                bucket_name=obj.bucket_name,
                object_name=obj.object_name
            )

    def recreate_bucket(self):
        self.client.remove_bucket(bucket_name=self.config.bucket)
        self.ensure_bucket(bucket_name=self.config.bucket)



from setuptools import setup, find_packages


with open('README.md') as f:
    long_description = ''.join(f.readlines())

setup(
    name='dsw2to3',
    version='0.1.4',
    description='CLI tool to support migration from DSW 2.X to DSW 3.0',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Marek Suchánek',
    keywords='dsw migration database upgrade',
    license='Apache License 2.0',
    url='https://github.com/ds-wizard/dsw2to3',
    packages=find_packages(),
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Topic :: Database',
        'Topic :: Utilities',
    ],
    zip_safe=False,
    python_requires='>=3.6, <4',
    install_requires=[
        'click',
        'minio',
        'pymongo',
        'PyYAML',
        'psycopg2',
        'tenacity',
    ],
    entry_points={
        'console_scripts': [
            'dsw2to3=dsw2to3:main',
        ],
    },
)



import logging
import sys

from dsw2to3.logger import LOGGER


class MigrationError(RuntimeError):

    def __init__(self, cause: str, message: str, level: int):
        super().__init__()
        self.cause = cause
        self.message = message
        self.level = level


def _log_handle(error: MigrationError):
    LOGGER.log(level=error.level, msg=f'{error.cause}:  {error.message}')


def _stop_handle(error: MigrationError):
    if error.level >= logging.ERROR:
        LOGGER.critical(msg=error.message)
        LOGGER.info('Exiting... you can try to run with --best-effort flag')
        sys.exit(1)
    else:
        _log_handle(error=error)


class MigrationErrorHandler:

    def __init__(self):
        self._handler = _stop_handle

    def set_log(self):
        self._handler = _log_handle

    def set_stop(self):
        self._handler = _stop_handle

    def warning(self, cause: str, message: str):
        self._handler(MigrationError(cause=cause, message=message, level=logging.WARNING))

    def error(self, cause: str, message: str):
        self._handler(MigrationError(cause=cause, message=message, level=logging.ERROR))

    def critical(self, cause: str, message: str):
        self._handler(MigrationError(cause=cause, message=message, level=logging.CRITICAL))


ERROR_HANDLER = MigrationErrorHandler()



import dataclasses
import yaml

from psycopg2.extensions import parse_dsn
from typing import Optional


class MissingConfigurationError(Exception):

    def __init__(self, missing: list[str]):
        self.missing = missing


@dataclasses.dataclass
class MongoConfig:
    host: str
    port: int
    database: str
    auth_enabled: bool
    username: Optional[str]
    password: Optional[str]

    @property
    def mongo_client_kwargs(self):
        kwargs = {
            'host': self.host, 'port': self.port
        }
        if self.auth_enabled:
            kwargs['username'] = self.username
            kwargs['password'] = self.password
            kwargs['authSource'] = self.database
        return kwargs

    def __str__(self):
        return f'{self.host}:{self.port}/{self.database}'


@dataclasses.dataclass
class PostgresConfig:
    connection_string: str


@dataclasses.dataclass
class S3Config:
    url: str
    username: str
    password: str
    bucket: str


@dataclasses.dataclass
class Config:
    mongo: MongoConfig
    postgres: PostgresConfig
    s3: S3Config


class ConfigParser:
    DEFAULTS = {
        'database': {
            'host': 'mongo',
            'port': 27017,
            'authEnabled': False,
            'username': None,
            'password': None,
        },
        's3': {
            'url': 'http://minio:9000',
            'bucket': 'engine-wizard',
            'username': 'minio',
            'password': 'minioPassword'
        }
    }
    REQUIRED = [
        ['database', 'connectionString'],
    ]

    def __init__(self):
        self.cfg = dict()

    @staticmethod
    def can_read(content):
        try:
            yaml.load(content, Loader=yaml.FullLoader)
            return True
        except Exception:
            return False

    def read_file(self, fp):
        self.cfg = yaml.load(fp, Loader=yaml.FullLoader)

    def read_string(self, content):
        self.cfg = yaml.load(content, Loader=yaml.FullLoader)

    def has(self, *path):
        x = self.cfg
        for p in path:
            if not hasattr(x, 'keys') or p not in x.keys():
                return False
            x = x[p]
        return True

    def _get_default(self, *path):
        x = self.DEFAULTS
        for p in path:
            x = x[p]
        return x

    def get_or_default(self, *path):
        x = self.cfg
        for p in path:
            if not hasattr(x, 'keys') or p not in x.keys():
                return self._get_default(*path)
            x = x[p]
        return x

    def validate(self):
        missing = []
        for path in self.REQUIRED:
            if not self.has(*path):
                missing.append('.'.join(path))
        if len(missing) > 0:
            raise MissingConfigurationError(missing)

    @property
    def mongo(self):
        return MongoConfig(
            host=self.get_or_default('database', 'host'),
            port=self.get_or_default('database', 'port'),
            database=self.get_or_default('database', 'databaseName'),
            auth_enabled=self.get_or_default('database', 'authEnabled'),
            username=self.get_or_default('database', 'username'),
            password=self.get_or_default('database', 'password'),
        )

    @property
    def postgres(self):
        connection_string = self.get_or_default('database', 'connectionString')
        parsed = parse_dsn(connection_string)
        return PostgresConfig(
            connection_string=connection_string,
        )

    @property
    def s3(self):
        return S3Config(
            url=self.get_or_default('s3', 'url'),
            bucket=self.get_or_default('s3', 'bucket'),
            username=self.get_or_default('s3', 'username'),
            password=self.get_or_default('s3', 'password'),
        )

    @property
    def config(self):
        return Config(
            mongo=self.mongo,
            postgres=self.postgres,
            s3=self.s3,
        )



import dataclasses
import datetime
import json
import uuid

from typing import Optional

from dsw2to3.migration.common import insert_query
from dsw2to3.logger import LOGGER


class DSWJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            ts = obj.isoformat(timespec='microseconds')
            return f'{ts}Z'
        return super().default(obj)


def _wrap_json(data):
    return json.dumps(data, cls=DSWJsonEncoder)


@dataclasses.dataclass
class ACLGroup:
    TABLE_NAME = 'acl_group'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'id',
            'name',
            'description',
        ]
    )

    id: str  # varchar PK
    name: str  # varchar
    description: str  # varchar
    integrity_ok: bool = True

    def get_id(self):
        return self.id

    def query_vars(self):
        return (
            self.id,
            self.name,
            self.description,
        )


@dataclasses.dataclass
class ActionKey:
    COLLECTION = 'actionKeys'
    TABLE_NAME = 'action_key'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'uuid',
            'user_id',
            'type',
            'hash',
            'created_at',
        ]
    )

    uuid: str  # uuid PK
    user_id: str  # uuid FK??
    type: str  # varchar
    hash: str  # varchar
    created_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def get_id(self):
        return self.uuid

    def query_vars(self):
        return (
            self.uuid,
            self.user_id,
            self.type,
            self.hash,
            self.created_at,
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return ActionKey(
            uuid=doc.get('uuid'),
            type=doc.get('type'),
            hash=doc.get('hash'),
            user_id=doc.get('userId'),
            created_at=doc.get('createdAt', now),
        )


@dataclasses.dataclass
class AppConfig:
    COLLECTION = 'appConfigs'
    TABLE_NAME = 'app_config'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'id',
            'organization',
            'authentication',
            'privacy_and_support',
            'dashboard',
            'look_and_feel',
            'registry',
            'knowledge_model',
            'questionnaire',
            'template',
            'submission',
            'created_at',
            'updated_at',
        ]
    )

    _DEFAULT_KM = {
        'public': {
            'enabled': False,
            'packages': [],
        },
    }

    # id: int  # bigint PK
    organization: dict  # json
    authentication: dict  # json
    privacy_and_support: dict  # json
    dashboard: dict  # json
    look_and_feel: dict  # json
    registry: dict  # json
    knowledge_model: dict  # json
    questionnaire: dict  # json
    template: dict  # json
    submission: dict  # json
    created_at: datetime.datetime  # timestamp+tz
    updated_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def get_id(self):
        return 1

    def query_vars(self):
        return (
            self.get_id(),
            _wrap_json(self.organization),
            _wrap_json(self.authentication),
            _wrap_json(self.privacy_and_support),
            _wrap_json(self.dashboard),
            _wrap_json(self.look_and_feel),
            _wrap_json(self.registry),
            _wrap_json(self.knowledge_model),
            _wrap_json(self.questionnaire),
            _wrap_json(self.template),
            _wrap_json(self.submission),
            self.created_at,
            self.updated_at,
        )

    def _migrate(self):
        self.questionnaire['questionnaireSharing']['anonymousEnabled'] = False

    @classmethod
    def from_mongo(cls, doc: dict, now: datetime.datetime):
        app_config = AppConfig(
            organization=doc.get('organization'),
            authentication=doc.get('authentication'),
            privacy_and_support=doc.get('privacyAndSupport'),
            dashboard=doc.get('dashboard'),
            look_and_feel=doc.get('lookAndFeel'),
            registry=doc.get('registry'),
            knowledge_model=cls._DEFAULT_KM,
            questionnaire=doc.get('questionnaire'),
            template=doc.get('template'),
            submission=doc.get('submission'),
            created_at=doc.get('createdAt', now),
            updated_at=doc.get('updatedAt', now),
        )
        app_config._migrate()
        return app_config


@dataclasses.dataclass
class BookReference:
    COLLECTION = 'bookReferences'
    TABLE_NAME = 'book_reference'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'short_uuid',
            'book_chapter',
            'content',
            'created_at',
            'updated_at',
        ]
    )

    short_uuid: str  # varchar PK
    book_chapter: str  # varchar
    content: str  # varchar
    created_at: datetime.datetime  # timestamp+tz
    updated_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def query_vars(self):
        return (
            self.short_uuid,
            self.book_chapter,
            self.content,
            self.created_at,
            self.updated_at,
        )

    def get_id(self):
        return self.short_uuid

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return BookReference(
            short_uuid=doc.get('shortUuid'),
            book_chapter=doc.get('bookChapter'),
            content=doc.get('content'),
            created_at=doc.get('createdAt', now),
            updated_at=doc.get('updatedAt', now),
        )


@dataclasses.dataclass
class Branch:
    COLLECTION = 'branches'
    TABLE_NAME = 'branch'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'uuid',
            'name',
            'km_id',
            'metamodel_version',
            'previous_package_id',
            'events',
            'owner_uuid',
            'created_at',
            'updated_at',
        ]
    )

    uuid: str  # uuid PK
    name: str  # varchar
    km_id: str  # varchar
    metamodel_version: int  # integer
    previous_package_id: str  # varchar
    events: list  # json
    owner_uuid: str  # uuid FK??
    created_at: datetime.datetime  # timestamp+tz
    updated_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def get_id(self):
        return self.uuid

    def query_vars(self):
        return (
            self.uuid,
            self.name,
            self.km_id,
            self.metamodel_version,
            self.previous_package_id,
            _wrap_json(self.events),
            self.owner_uuid,
            self.created_at,
            self.updated_at,
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return Branch(
            uuid=doc.get('uuid'),
            name=doc.get('name'),
            km_id=doc.get('kmId'),
            metamodel_version=doc.get('metamodelVersion'),
            previous_package_id=doc.get('previousPackageId'),
            events=doc.get('events'),
            owner_uuid=doc.get('ownerUuid'),
            created_at=doc.get('createdAt', now),
            updated_at=doc.get('updatedAt', now),
        )


@dataclasses.dataclass
class Document:
    COLLECTION = 'documents'
    TABLE_NAME = 'document'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'uuid',
            'name',
            'state',
            'durability',
            'questionnaire_uuid',
            'questionnaire_event_uuid',
            'questionnaire_replies_hash',
            'template_id',
            'format_uuid',
            'metadata',
            'creator_uuid',
            'retrieved_at',
            'finished_at',
            'created_at',
        ]
    )

    uuid: str  # uuid PK
    name: str  # varchar
    state: str  # varchar
    durability: str  # varchar
    questionnaire_uuid: str  # uuid FK??
    questionnaire_event_uuid: str  # uuid
    questionnaire_replies_hash: int  # bigint
    template_id: str  # varchar FK??
    format_uuid: str  # uuid
    metadata: dict  # json
    creator_uuid: str  # uuid FK??
    retrieved_at: Optional[datetime.datetime]  # timestamp+tz
    finished_at: Optional[datetime.datetime]  # timestamp+tz
    created_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def get_id(self):
        return self.uuid

    def query_vars(self):
        return (
            self.uuid,
            self.name,
            self.state,
            self.durability,
            self.questionnaire_uuid,
            self.questionnaire_event_uuid,
            self.questionnaire_replies_hash,
            self.template_id,
            self.format_uuid,
            _wrap_json(self.metadata),
            self.creator_uuid,
            self.retrieved_at,
            self.finished_at,
            self.created_at,
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return Document(
            uuid=doc.get('uuid'),
            name=doc.get('name'),
            state=doc.get('state'),
            durability=doc.get('durability'),
            questionnaire_uuid=doc.get('questionnaireUuid'),
            questionnaire_event_uuid=doc.get('questionnaireEventUuid'),
            questionnaire_replies_hash=doc.get('questionnaireRepliesHash'),
            template_id=doc.get('templateId'),
            format_uuid=doc.get('formatUuid'),
            metadata=doc.get('metadata'),
            creator_uuid=doc.get('creatorUuid'),
            retrieved_at=doc.get('retrievedAt', None),
            finished_at=doc.get('finishedAt', None),
            created_at=doc.get('createdAt', now),
        )


@dataclasses.dataclass
class Feedback:
    COLLECTION = 'feedbacks'
    TABLE_NAME = 'feedback'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'uuid',
            'issue_id',
            'question_uuid',
            'package_id',
            'title',
            'content',
            'created_at',
            'updated_at',
        ]
    )

    uuid: str  # uuid PK
    issue_id: int  # integer
    question_uuid: str  # uuid
    package_id: str  # varchar FK??
    title: str  # varchar
    content: str  # varchar
    created_at: datetime.datetime  # timestamp+tz
    updated_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def get_id(self):
        return self.uuid

    def query_vars(self):
        return (
            self.uuid,
            self.issue_id,
            self.question_uuid,
            self.package_id,
            self.title,
            self.content,
            self.created_at,
            self.updated_at,
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return Feedback(
            uuid=doc.get('uuid'),
            issue_id=doc.get('issueId'),
            question_uuid=doc.get('questionUuid'),
            package_id=doc.get('packageId'),
            title=doc.get('title'),
            content=doc.get('content'),
            created_at=doc.get('createdAt', now),
            updated_at=doc.get('updatedAt', now),
        )


@dataclasses.dataclass
class KMMigration:
    COLLECTION = 'kmMigrations'
    TABLE_NAME = 'knowledge_model_migration'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'branch_uuid',
            'metamodel_version',
            'migration_state',
            'branch_previous_package_id',
            'target_package_id',
            'branch_events',
            'target_package_events',
            'result_events',
            'current_knowledge_model',
        ]
    )

    branch_uuid: str  # uuid PK
    metamodel_version: int  # integer
    migration_state: dict  # json
    branch_previous_package_id: str  # varchar FK??
    target_package_id: str  # varchar FK??
    branch_events: list  # json
    target_package_events: list  # json
    result_events: list  # json
    current_knowledge_model: dict  # json
    integrity_ok: bool = True

    def get_id(self):
        return self.branch_uuid

    def query_vars(self):
        return (
            self.branch_uuid,
            self.metamodel_version,
            _wrap_json(self.migration_state),
            self.branch_previous_package_id,
            self.target_package_id,
            _wrap_json(self.branch_events),
            _wrap_json(self.target_package_events),
            _wrap_json(self.result_events),
            _wrap_json(self.current_knowledge_model),
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return KMMigration(
            branch_uuid=doc.get('branchUuid'),
            metamodel_version=doc.get('metamodelVersion'),
            migration_state=doc.get('migrationState'),
            branch_previous_package_id=doc.get('branchPreviousPackageId'),
            target_package_id=doc.get('targetPackageId'),
            branch_events=doc.get('branchEvents'),
            target_package_events=doc.get('targetPackageEvents'),
            result_events=doc.get('resultEvents'),
            current_knowledge_model=doc.get('currentKnowledgeModel'),
        )


@dataclasses.dataclass
class Level:
    COLLECTION = 'levels'
    TABLE_NAME = 'level'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'level',
            'title',
            'description',
            'created_at',
            'updated_at',
        ]
    )

    level: int  # integer PK
    title: str  # varchar
    description: str  # varchar
    created_at: datetime.datetime  # timestamp+tz
    updated_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def get_id(self):
        return self.level

    def query_vars(self):
        return (
            self.level,
            self.title,
            self.description,
            self.created_at,
            self.updated_at,
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return Level(
            level=doc.get('level'),
            title=doc.get('title'),
            description=doc.get('description'),
            created_at=doc.get('createdAt', now),
            updated_at=doc.get('updatedAt', now),
        )


@dataclasses.dataclass
class Metric:
    COLLECTION = 'metrics'
    TABLE_NAME = 'metric'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'uuid',
            'title',
            'abbreviation',
            'description',
            'reference_json',
            'created_at',
            'updated_at',
        ]
    )

    uuid: int  # uuid PK
    title: str  # varchar
    abbreviation: str  # varchar
    description: str  # varchar
    reference_json: dict  # json
    created_at: datetime.datetime  # timestamp+tz
    updated_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def get_id(self):
        return self.uuid

    def query_vars(self):
        return (
            self.uuid,
            self.title,
            self.abbreviation,
            self.description,
            _wrap_json(self.reference_json),
            self.created_at,
            self.updated_at,
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return Metric(
            uuid=doc.get('uuid'),
            title=doc.get('title'),
            abbreviation=doc.get('abbreviation'),
            description=doc.get('description'),
            reference_json=doc.get('references'),
            created_at=doc.get('createdAt', now),
            updated_at=doc.get('updatedAt', now),
        )


@dataclasses.dataclass
class Package:
    COLLECTION = 'packages'
    TABLE_NAME = 'package'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'id',
            'name',
            'organization_id',
            'km_id',
            'version',
            'metamodel_version',
            'description',
            'readme',
            'license',
            'previous_package_id',
            'fork_of_package_id',
            'merge_checkpoint_package_id',
            'events',
            'created_at',
        ]
    )

    id: str  # varchar PK
    name: str  # varchar
    organization_id: str  # varchar
    km_id: str  # varchar
    version: str  # varchar
    metamodel_version: int  # integer
    description: str  # varchar
    readme: str  # varchar
    license: str  # varchar
    previous_package_id: str  # varchar FK
    fork_of_package_id: str  # varchar FK
    merge_checkpoint_package_id: str  # varchar FK
    events: list  # json
    created_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def get_id(self):
        return self.id

    def query_vars(self):
        return (
            self.id,
            self.name,
            self.organization_id,
            self.km_id,
            self.version,
            self.metamodel_version,
            self.description,
            self.readme,
            self.license,
            self.previous_package_id,
            self.fork_of_package_id,
            self.merge_checkpoint_package_id,
            _wrap_json(self.events),
            self.created_at,
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return Package(
            id=doc.get('id'),
            name=doc.get('name'),
            organization_id=doc.get('organizationId'),
            km_id=doc.get('kmId'),
            version=doc.get('version'),
            metamodel_version=doc.get('metamodelVersion'),
            description=doc.get('description'),
            readme=doc.get('readme'),
            license=doc.get('license'),
            previous_package_id=doc.get('previousPackageId'),
            fork_of_package_id=doc.get('forkOfPackageId'),
            merge_checkpoint_package_id=doc.get('mergeCheckpointPackageId'),
            events=doc.get('events'),
            created_at=doc.get('createdAt', now),
        )


@dataclasses.dataclass
class QuestionnaireMigration:
    COLLECTION = 'questionnaireMigrations'
    TABLE_NAME = 'questionnaire_migration'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'old_questionnaire_uuid',
            'new_questionnaire_uuid',
            'resolved_question_uuids',
        ]
    )

    old_questionnaire_uuid: str  # uuid PK FK??
    new_questionnaire_uuid: str  # uuid PK FK??
    resolved_question_uuids: list  # json
    integrity_ok: bool = True

    def get_id(self):
        return f'{self.old_questionnaire_uuid}_{self.new_questionnaire_uuid}'

    def query_vars(self):
        return (
            self.old_questionnaire_uuid,
            self.new_questionnaire_uuid,
            _wrap_json(self.resolved_question_uuids),
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return QuestionnaireMigration(
            old_questionnaire_uuid=doc.get('oldQuestionnaireUuid'),
            new_questionnaire_uuid=doc.get('newQuestionnaireUuid'),
            resolved_question_uuids=doc.get('resolvedQuestionUuids', []),
        )


@dataclasses.dataclass
class Questionnaire:
    COLLECTION = 'questionnaires'
    TABLE_NAME = 'questionnaire'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'uuid',
            'name',
            'visibility',
            'sharing',
            'package_id',
            'selected_tag_uuids',
            'template_id',
            'format_uuid',
            'creator_uuid',
            'events',
            'versions',
            'created_at',
            'updated_at',
        ]
    )

    uuid: str  # uuid PK
    name: str  # varchar
    visibility: str  # varchar
    sharing: str  # varchar
    package_id: str  # varchar FK??
    selected_tag_uuids: list  # json
    template_id: str  # varchar FK??
    format_uuid: str  # uuid
    creator_uuid: str  # uuid FK??
    events: list  # json
    versions: list  # json
    created_at: datetime.datetime  # timestamp+tz
    updated_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def get_id(self):
        return self.uuid

    def query_vars(self):
        return (
            self.uuid,
            self.name,
            self.visibility,
            self.sharing,
            self.package_id,
            _wrap_json(self.selected_tag_uuids),
            self.template_id,
            self.format_uuid,
            self.creator_uuid,
            _wrap_json(self.events),
            _wrap_json(self.versions),
            self.created_at,
            self.updated_at,
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return Questionnaire(
            uuid=doc.get('uuid'),
            name=doc.get('name'),
            visibility=doc.get('visibility'),
            sharing=doc.get('sharing'),
            package_id=doc.get('packageId'),
            selected_tag_uuids=doc.get('selectedTagUuids', []),
            template_id=doc.get('templateId', None),
            format_uuid=doc.get('formatUuid', None),
            creator_uuid=doc.get('creatorUuid', None),
            events=doc.get('events', []),
            versions=doc.get('versions', []),
            created_at=doc.get('createdAt', now),
            updated_at=doc.get('updatedAt', now),
        )


@dataclasses.dataclass
class QuestionnaireACLGroup:
    TABLE_NAME = 'questionnaire_acl_group'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'uuid',
            'group_id',
            'perms',
            'questionnaire_uuid',
        ]
    )

    uuid: str  # uuid PK
    group_id: str  # uuid FK
    perms: list  # text[]
    questionnaire_uuid: str  # uuid FK
    integrity_ok: bool = True

    def get_id(self):
        return self.uuid

    def query_vars(self):
        return (
            self.uuid,
            self.group_id,
            self.perms,
            self.questionnaire_uuid,
        )


@dataclasses.dataclass
class QuestionnaireACLUser:
    TABLE_NAME = 'questionnaire_acl_user'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'uuid',
            'user_uuid',
            'perms',
            'questionnaire_uuid',
        ]
    )

    uuid: str  # uuid PK
    user_uuid: str  # uuid FK
    perms: list  # text[]
    questionnaire_uuid: str  # uuid FK
    integrity_ok: bool = True

    def get_id(self):
        return self.uuid

    def query_vars(self):
        return (
            self.uuid,
            self.user_uuid,
            self.perms,
            self.questionnaire_uuid,
        )

    @staticmethod
    def from_mongo(doc: dict, permission: dict, now: datetime.datetime):
        return QuestionnaireACLUser(
            uuid=str(uuid.uuid4()),
            user_uuid=permission.get('member').get('uuid'),
            perms=permission.get('perms'),
            questionnaire_uuid=doc.get('uuid'),
        )


@dataclasses.dataclass
class Template:
    COLLECTION = 'templates'
    TABLE_NAME = 'template'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'id',
            'name',
            'organization_id',
            'template_id',
            'version',
            'metamodel_version',
            'description',
            'readme',
            'license',
            'allowed_packages',
            'recommended_package_id',
            'formats',
            'created_at',
        ]
    )

    id: str  # varchar PK
    name: str  # varchar
    organization_id: str  # varchar
    template_id: str  # varchar
    version: str  # varchar
    metamodel_version: int  # integer
    description: str  # varchar
    readme: str  # varchar
    license: str  # varchar
    allowed_packages: list  # json
    recommended_package_id: str  # varchar
    formats: list  # json
    created_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def get_id(self):
        return self.id

    def query_vars(self):
        return (
            self.id,
            self.name,
            self.organization_id,
            self.template_id,
            self.version,
            self.metamodel_version,
            self.description,
            self.readme,
            self.license,
            _wrap_json(self.allowed_packages),
            self.recommended_package_id,
            _wrap_json(self.formats),
            self.created_at,
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return Template(
            id=doc.get('id'),
            name=doc.get('name'),
            organization_id=doc.get('organizationId'),
            template_id=doc.get('templateId'),
            version=doc.get('version'),
            metamodel_version=doc.get('metamodelVersion'),
            description=doc.get('description'),
            readme=doc.get('readme'),
            license=doc.get('license'),
            allowed_packages=doc.get('allowedPackages'),
            recommended_package_id=doc.get('recommendedPackageId'),
            formats=doc.get('formats'),
            created_at=doc.get('createdAt', now),
        )


@dataclasses.dataclass
class TemplateAsset:
    TABLE_NAME = 'template_asset'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'template_id',
            'uuid',
            'file_name',
            'content_type',
        ]
    )

    template_id: str  # varchar FK
    uuid: str  # uuid PK
    file_name: str  # varchar
    content_type: str  # varchar
    original_uuid: str
    integrity_ok: bool = True

    def get_id(self):
        return self.uuid

    def query_vars(self):
        return (
            self.template_id,
            self.uuid,
            self.file_name,
            self.content_type,
        )

    @staticmethod
    def from_mongo(doc: dict, asset: dict, now: datetime.datetime):
        return TemplateAsset(
            template_id=doc.get('id'),
            uuid=str(uuid.uuid4()),
            file_name=asset.get('fileName'),
            content_type=asset.get('contentType'),
            original_uuid=asset.get('uuid'),
        )


@dataclasses.dataclass
class TemplateFile:
    TABLE_NAME = 'template_file'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'template_id',
            'uuid',
            'file_name',
            'content',
        ]
    )

    template_id: str  # varchar FK
    uuid: str  # uuid PK
    file_name: str  # varchar
    content: str  # varchar
    original_uuid: str
    integrity_ok: bool = True

    def get_id(self):
        return self.uuid

    def query_vars(self):
        return (
            self.template_id,
            self.uuid,
            self.file_name,
            self.content,
        )

    @staticmethod
    def from_mongo(doc: dict, file: dict, now: datetime.datetime):
        return TemplateFile(
            template_id=doc.get('id'),
            uuid=str(uuid.uuid4()),
            file_name=file.get('fileName'),
            content=file.get('content'),
            original_uuid=file.get('uuid'),
        )


@dataclasses.dataclass
class User:
    COLLECTION = 'users'
    TABLE_NAME = 'user_entity'
    INSERT_QUERY = insert_query(
        table_name=TABLE_NAME,
        fields=[
            'uuid',
            'first_name',
            'last_name',
            'email',
            'password_hash',
            'affiliation',
            'sources',
            'role',
            'permissions',
            'active',
            'submissions_props',
            'image_url',
            'groups',
            'last_visited_at',
            'created_at',
            'updated_at',
        ]
    )

    uuid: str  # uuid PK
    first_name: str  # varchar
    last_name: str  # varchar
    email: str  # varchar
    password_hash: str  # varchar
    affiliation: str  # varchar
    sources: list  # json
    role: str  # varchar
    permissions: list  # text[]
    active: bool  # boolean
    submissions_props: list  # json
    image_url: str  # varchar
    groups: list  # json
    last_visited_at: datetime.datetime  # timestamp+tz
    created_at: datetime.datetime  # timestamp+tz
    updated_at: datetime.datetime  # timestamp+tz
    integrity_ok: bool = True

    def get_id(self):
        return self.uuid

    def query_vars(self):
        return (
            self.uuid,
            self.first_name,
            self.last_name,
            self.email,
            self.password_hash,
            self.affiliation,
            _wrap_json(self.sources),
            self.role,
            self.permissions,
            self.active,
            _wrap_json(self.submissions_props),
            self.image_url,
            _wrap_json(self.groups),
            self.last_visited_at,
            self.created_at,
            self.updated_at,
        )

    @staticmethod
    def from_mongo(doc: dict, now: datetime.datetime):
        return User(
            uuid=doc.get('uuid'),
            first_name=doc.get('firstName'),
            last_name=doc.get('lastName'),
            email=doc.get('email'),
            password_hash=doc.get('passwordHash'),
            affiliation=doc.get('affiliation', None),
            sources=doc.get('sources'),
            role=doc.get('role'),
            permissions=doc.get('permissions'),
            active=doc.get('active', False),
            submissions_props=doc.get('submissionProps', []),
            image_url=doc.get('imageUrl', None),
            groups=doc.get('groups', []),
            last_visited_at=doc.get('lastVisitedAt', now),
            created_at=doc.get('createdAt', now),
            updated_at=doc.get('updatedAt', now),
        )


@dataclasses.dataclass
class WizardEntities:
    acl_groups: list[ACLGroup] = dataclasses.field(default_factory=list)
    action_keys: list[ActionKey] = dataclasses.field(default_factory=list)
    app_configs: list[AppConfig] = dataclasses.field(default_factory=list)
    book_references: list[BookReference] = dataclasses.field(default_factory=list)
    branches: list[Branch] = dataclasses.field(default_factory=list)
    documents: list[Document] = dataclasses.field(default_factory=list)
    feedbacks: list[Feedback] = dataclasses.field(default_factory=list)
    km_migrations: list[KMMigration] = dataclasses.field(default_factory=list)
    levels: list[Level] = dataclasses.field(default_factory=list)
    metrics: list[Metric] = dataclasses.field(default_factory=list)
    packages: list[Package] = dataclasses.field(default_factory=list)
    questionnaire_migrations: list[QuestionnaireMigration] = dataclasses.field(default_factory=list)
    questionnaires: list[Questionnaire] = dataclasses.field(default_factory=list)
    questionnaire_acl_groups: list[QuestionnaireACLGroup] = dataclasses.field(default_factory=list)
    questionnaire_acl_users: list[QuestionnaireACLUser] = dataclasses.field(default_factory=list)
    templates: list[Template] = dataclasses.field(default_factory=list)
    template_assets: list[TemplateAsset] = dataclasses.field(default_factory=list)
    template_files: list[TemplateFile] = dataclasses.field(default_factory=list)
    users: list[User] = dataclasses.field(default_factory=list)

    _result: list[str] = dataclasses.field(default_factory=list)

    LIST_ENTITY = {
        'acl_groups': ACLGroup,
        'action_keys': ActionKey,
        'app_configs': AppConfig,
        'book_references': BookReference,
        'branches': Branch,
        'documents': Document,
        'feedbacks': Feedback,
        'km_migrations': KMMigration,
        'levels': Level,
        'metrics': Metric,
        'packages': Package,
        'questionnaire_migrations': QuestionnaireMigration,
        'questionnaires': Questionnaire,
        'questionnaire_acl_groups': QuestionnaireACLGroup,
        'questionnaire_acl_users': QuestionnaireACLUser,
        'templates': Template,
        'template_assets': TemplateAsset,
        'template_files': TemplateFile,
        'users': User,
    }
    ENTITIES = [e for e in LIST_ENTITY.values()]
    ENTITY_LIST = {e.__name__: lst for lst, e in LIST_ENTITY.items()}

    def list_by_entity(self, entity) -> list:
        return getattr(self, self.ENTITY_LIST.get(entity.__name__))

    def set_list(self, entity, lst: list):
        return setattr(self, self.ENTITY_LIST.get(entity.__name__), lst)

    def valid_entries(self, entity):
        return (x for x in self.list_by_entity(entity) if x.integrity_ok is not False)

    def clear(self):
        for entity in self.ENTITIES:
            self.list_by_entity(entity).clear()

    def _check_uniqueness(self, list_name: str, entity):
        unique_set = set()
        for item in getattr(self, list_name):
            item_id = item.get_id()
            if item_id in unique_set:
                item.integrity_ok = False
                self._result.append(
                    f'Duplicate ID for {entity.__name__}: {item_id}'
                )
            unique_set.add(item_id)
        unique_set.clear()

    def _make_ids_set(self, entity) -> frozenset:
        return frozenset((
            item.get_id() for item in self.valid_entries(entity)
        ))

    def _check_reference(self, entity, target_entity, field_name: str, ids_set: frozenset):
        for item in self.valid_entries(entity):
            ref_id = getattr(item, field_name)
            if ref_id not in ids_set:
                item.integrity_ok = False
                self._result.append(
                    f'Missing {target_entity.__name__} ({field_name}={ref_id}) '
                    f'for {entity.__name__}: {item.get_id()}'
                )

    def _check_optional_reference(self, entity, target_entity, field_name: str, ids_set: frozenset):
        for item in self.valid_entries(entity):
            ref_id = getattr(item, field_name)
            if ref_id is not None and ref_id not in ids_set:
                item.integrity_ok = False
                self._result.append(
                    f'Missing {target_entity.__name__} ({field_name}={ref_id}) '
                    f'for {entity.__name__}: {item.get_id()}'
                )

    def _check_references(self) -> bool:
        prev_len = len(self._result)
        LOGGER.debug('  - collecting IDs of consistent entries')
        user_ids = self._make_ids_set(User)
        package_ids = self._make_ids_set(Package)
        template_ids = self._make_ids_set(Template)
        questionnaire_ids = self._make_ids_set(Questionnaire)
        branch_ids = self._make_ids_set(Branch)
        group_ids = self._make_ids_set(ACLGroup)

        LOGGER.debug('  - checking references')
        self._check_reference(ActionKey, User, 'user_id', user_ids)
        self._check_optional_reference(Branch, Package, 'previous_package_id', package_ids)
        self._check_reference(Branch, User, 'owner_uuid', user_ids)
        self._check_reference(Document, Questionnaire, 'questionnaire_uuid', questionnaire_ids)
        self._check_reference(Document, Template, 'template_id', template_ids)
        self._check_reference(Document, User, 'creator_uuid', user_ids)
        self._check_reference(Feedback, Package, 'package_id', package_ids)
        self._check_reference(KMMigration, Package, 'branch_previous_package_id', package_ids)
        self._check_reference(KMMigration, Package, 'target_package_id', package_ids)
        self._check_reference(KMMigration, Branch, 'branch_uuid', branch_ids)
        self._check_optional_reference(Package, Package, 'previous_package_id', package_ids)
        self._check_optional_reference(Package, Package, 'fork_of_package_id', package_ids)
        self._check_optional_reference(Package, Package, 'merge_checkpoint_package_id', package_ids)
        self._check_reference(Questionnaire, Package, 'package_id', package_ids)
        self._check_optional_reference(Questionnaire, Template, 'template_id', template_ids)
        self._check_optional_reference(Questionnaire, User, 'creator_uuid', user_ids)
        self._check_reference(QuestionnaireACLGroup, ACLGroup, 'group_id', group_ids)
        self._check_reference(QuestionnaireACLGroup, Questionnaire, 'questionnaire_uuid', questionnaire_ids)
        self._check_reference(QuestionnaireACLUser, Questionnaire, 'questionnaire_uuid', questionnaire_ids)
        self._check_reference(QuestionnaireACLUser, User, 'user_uuid', user_ids)
        self._check_reference(TemplateAsset, Template, 'template_id', template_ids)
        self._check_reference(TemplateFile, Template, 'template_id', template_ids)

        LOGGER.debug(f'  - {len(self._result) - prev_len} new inconsistencies found')
        return prev_len != len(self._result)

    def check_integrity(self) -> list[str]:
        self._result.clear()
        # Uniqueness (PKs, UNIQUE)
        LOGGER.info('- checking ID uniqueness')
        for lst, e in self.LIST_ENTITY.items():
            self._check_uniqueness(list_name=lst, entity=e)
        # References (FKs)
        LOGGER.info('- checking references for the first time')
        check_again = self._check_references()
        i = 1
        while check_again:
            LOGGER.info(f'- checking references again (#{i})')
            i += 1
            check_again = self._check_references()
        return self._result
